/*
** $Id: ComposeFrame.java,v 1.4 2002/03/25 02:51:29 mcs Exp $
**
** Copyright (c) 2000-2002 Jeff Gay
** on behalf of ICEMail.org <http://www.icemail.org>
** Copyright (c) 1998-2000 by Timothy Gerard Endres
** 
** This program is free software.
** 
** You may redistribute it and/or modify it under the terms of the GNU
** General Public License as published by the Free Software Foundation.
** Version 2 of the license should be included with this distribution in
** the file LICENSE, as well as License.html. If the license is not
** included with this distribution, you may find a copy at the FSF web
** site at 'www.gnu.org' or 'www.fsf.org', or you may write to the
** Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139 USA.
**
** THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND,
** NOT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR
** OF THIS SOFTWARE, ASSUMES _NO_ RESPONSIBILITY FOR ANY
** CONSEQUENCE RESULTING FROM THE USE, MODIFICATION, OR
** REDISTRIBUTION OF THIS SOFTWARE. 
*/

package org.icemail.mail;

import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.FileDialog;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowAdapter;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;

import java.text.FieldPosition;
import java.text.MessageFormat;
import java.text.SimpleDateFormat;

import java.util.Date;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.MissingResourceException;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.activation.FileTypeMap;

import javax.mail.Address;
import javax.mail.Folder;
import javax.mail.Header;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Part;
import javax.mail.Store;

import javax.mail.internet.InternetAddress;
import javax.mail.internet.InternetHeaders;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.KeyStroke;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.SwingUtilities;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;

import javax.swing.text.Document;
import javax.swing.text.DefaultStyledDocument;
import javax.swing.text.JTextComponent;

import javax.swing.undo.UndoManager;

import org.icemail.Package;

import org.icemail.addressbook.AddressFieldActionListener;

import org.icemail.smime.SMIMELibrary;

import org.icemail.smime.MissingCertificateException;
import org.icemail.smime.PasswordCancelException;

import org.icemail.util.AWTUtilities;
import org.icemail.util.ComponentFactory;
import org.icemail.util.FormatResourceBundle;
import org.icemail.util.JFCUtilities;
import org.icemail.util.ResourceCache;
import org.icemail.util.RotateButton;
import org.icemail.util.StringUtilities;
import org.icemail.util.UserProperties;

// begin additions for LDAP lookup (mcs@netscape.com)
import netscape.ldap.*;
import java.net.MalformedURLException;
// end of additions by mcs@netscape.com

/**
 * This class composes and edits messages,
 * allowing messages to be sent, printed, and saved.
 * <p>
 * A message can contain:
 * <pre>
 * -sender and receipient addresses
 * -attachments
 * -extra headers
 * </pre>
 * <p>
 * Additionally, messages can be encrypted or signed when
 * the sMIME library is available.
 * <p>
 * At this time, JFrame is known to leak memory, so all ComposeFrames
 * are managed, and re-used.
 *
 * @see ComposeMgr
 */
public class ComposeFrame
  extends JFrame
{
  private static final int Debug_ = Package.DEBUG ? Package.getLevel( "ComposeFrame" ) : 0;

// state of mail editing
  public static final int CREATING = 0;
  public static final int EDITING = 1;
  public static final int REPLYING = 2;
  public static final int REPLYINGTOALL = 3;
  public static final int FORWARDING = 4;

  private static int[] AccKeys = {
    KeyEvent.VK_1, KeyEvent.VK_2, KeyEvent.VK_3, KeyEvent.VK_4,
    KeyEvent.VK_5, KeyEvent.VK_6, KeyEvent.VK_7, KeyEvent.VK_8,
    KeyEvent.VK_9, KeyEvent.VK_0,
    };

  private static String     LastAttachDir = null;

  private int               state_ = ComposeFrame.CREATING;
  private Message           originalMessage_ = null;

  private boolean           modified_ = false;

  private String            signature_;

  /**
   * The headers preserved from the original message in a forwarding.
   */
  private InternetHeaders   headers_;
  /**
   * The extra headers configured by the composer.
   */
  private InternetHeaders   extraHeaders_;

  /*
   * controls to maintain choice of character set
   */
  private Vector            charsets_;
  private JPopupMenu        charsetMenu_;
  private JLabel            charsetLbl_;

  private JMenuBar          menuBar_;

  private JTextField        fromText_;
  private JTextField        toText_;
  private JTextField        ccText_;
  private JTextField        bccText_;
  private JTextField        subjectText_;

  private JTextPane         messageText_;
  private UndoManager       messageUndo_;

  private JButton           sendButton_;
  private JButton           printButton_;
  private JButton           attachButton_;
  private JButton           detachButton_;
  private RotateButton      priorityButton_;
  private RotateButton      encryptButton_;
  private RotateButton      signButton_;

  private JComboBox         attachComboBox_;
  private Vector            attachments_;
  private int               textAttachmentSequence_ = 0;

  private IActionListener   actionListener_ = new IActionListener();
  private IFocusListener    focusListener_ = new IFocusListener();
  private JTextComponent    currentFocus_ = null;

// begin additions for LDAP lookup (mcs@netscape.com)
  private LDAPUrl			ldapDirectoryURL = null;
// end of additions by mcs@netscape.com

  /**
   * Default constructor, creating a composition frame for a new message.
   */
  public
  ComposeFrame() {
    super();

    setIconImage( ResourceCache.getImage( "/org/icemail/mail/images/icemail.gif" ) );

    if ( Package.DEBUG && Package.isTraceable( "ComposeFrame" ) ) {
      System.out.println( "ComposeFrame()" );
    }

    setDefaultCloseOperation( DO_NOTHING_ON_CLOSE );

    signature_ = null;
    headers_ = null;
    extraHeaders_ = null;
    attachments_ = null;
    currentFocus_ = null;

    charsets_ = Configuration.getCharsets();

    establishMenuBar();

    establishContents();

    addWindowListener( new IWindowAdapter() );

    pack();
    charsetMenu_.pack();

  //
  // FIXME try to select input method of selected locale
  // Removed until we decide to only support Java 2.
  //
  // getInputContext().selectInputMethod( locale );
  //
    loadLayoutProperties();
  }

  /**
   * Dispose of the window and release resources.
   * <p>
   * Implementation of java.awt.Window.dispose()
   */
  public void
  dispose( boolean automatically ) {
    if ( Package.DEBUG && Package.isTraceable( "ComposeFrame" ) ) {
      System.out.println( "ComposeFrame.dispose()" );
    }
    if ( automatically ) {
      dismissWindow();
    } else {
      closeWindow();
    }
  }

  /**
   * Edit the given message using the composition frame.
   *
   * @param state the state of the composition
   * @param msg the message to edit
   * @param body the optionally selected body of the message
   */
  public void
  editMessage( int state, Message msg, String body ) {
    if ( Package.DEBUG && Package.isTraceable( "ComposeFrame" ) ) {
      System.out.println( "ComposeFrame.editMessage(i,m,s): " + state );
    }

    setEnabled( true );
    state_ = state;
    originalMessage_ = msg;

    textAttachmentSequence_ = 0;
    signature_ = null;
    headers_ = new InternetHeaders();
    extraHeaders_ = new InternetHeaders();
    attachments_ = new Vector();
    currentFocus_ = null;

  // title the frame accordingly
    String rsrcName = "Compose.title.creating";
    switch ( state_ ) {
    case EDITING:
      rsrcName = "Compose.title.editing";
      break;
    case REPLYING:
      rsrcName = "Compose.title.replying";
      break;
    case REPLYINGTOALL:
      rsrcName = "Compose.title.replyingtoall";
      break;
    case FORWARDING:
      rsrcName = "Compose.title.forwarding";
      break;
    case CREATING:
      default:
      break;
    }
    setTitle( ICEMail.getBundle().getString( rsrcName ) );

  // clear previous controls
    if ( ICEMail.getSMIMELibrary() != null ) {
      priorityButton_.rotate( 0 );
      encryptButton_.rotate( 0 );
      signButton_.rotate( 0 );
    }

    setMessage( body );

    if ( state_ == REPLYING || state_ == REPLYINGTOALL ) {
      int end = messageText_.getDocument().getEndPosition().getOffset();
      messageText_.setCaretPosition( end - 1 );
      messageText_.moveCaretPosition( end - 1 );
    } else {
      messageText_.setCaretPosition( 0 );
      messageText_.moveCaretPosition( 0 );
    }

    modified_ = false;

    setVisible( true );
    toFront();

    if ( state_ == CREATING || state_ == FORWARDING ) {
      toText_.requestFocus();
    } else {
      messageText_.requestFocus();
    }
  }

//............................................................
// PRIVATE METHODS

  private void
  closeWindow() {
    if ( Package.DEBUG && Package.isTraceable( "ComposeFrame" ) ) {
      System.out.println( "ComposeFrame.closeWindow()" );
    }

    boolean xdismiss = true;

    String to = toText_.getText().trim();
    String cc = ccText_.getText().trim();
    String bcc = bccText_.getText().trim();
    String subject = subjectText_.getText().trim();

    if ( to.length() > 0  || cc.length() > 0 ||
         bcc.length() > 0 || subject.length() > 0 ) {
      modified_ = true;
    }

    if ( modified_ ) {
      int option =
        ComponentFactory.showDialog( ICEMail.getBundle(), "Compose.Save",
                                     JOptionPane.YES_NO_CANCEL_OPTION,
                                     JOptionPane.QUESTION_MESSAGE, null );

      if ( option == JOptionPane.YES_OPTION ) {
        try {
          MimeMessage xmessage = buildMessage();
          buildHeader( xmessage );
          xmessage.saveChanges();
          saveComposition( xmessage );
        } catch ( MessagingException xex ) {
          xdismiss = false;

          ByteArrayOutputStream xbaos = new ByteArrayOutputStream();
          xex.printStackTrace( new PrintWriter(  xbaos ) );

          Object[] xargs = new Object[2];
          xargs[0] = xex.getMessage();
          xargs[1] = xbaos.toString();
          ComponentFactory.showDialog( ICEMail.getBundle(), "Compose.SavingException",
                                       0, JOptionPane.ERROR_MESSAGE, xargs );
        }
      } else
      if ( option == JOptionPane.CANCEL_OPTION ) {
        xdismiss = false;
      }
    }

    if ( xdismiss ) {
      dismissWindow();
    }
  }

  private void
  dismissWindow() {
    saveLayoutProperties();

    originalMessage_ = null;
    signature_ = null;
    headers_ = null;
    extraHeaders_ = null;
    attachments_ = null;

    if ( attachComboBox_.getItemCount() > 0 ) {
      attachComboBox_.removeAllItems();
    }

    fromText_.setText( "" );
    toText_.setText( "" );
    ccText_.setText( "" );
    bccText_.setText( "" );
    subjectText_.setText( "" );

    messageText_.setText( "" );

  //  reasonable time to clean up memory.
    System.gc();

    javax.swing.SwingUtilities.invokeLater( new IWindowDisposer() );
  }

//............................................................

  private void
  loadLayoutProperties() {
    Dimension sz = getSize();

    Rectangle bounds = Configuration.getBounds( "composeWindow",
                                                new Rectangle( 20, 20, sz.width, sz.height ) );
    setBounds( bounds );

    charsetLbl_.setText( Configuration.getCharset() );
  }

  private void
  saveLayoutProperties() {
    Configuration.saveBounds( "composeWindow", getBounds() );
    Configuration.setCharset( charsetLbl_.getText() );
  }

  private Message
  buildHeader( MimeMessage message )
    throws MessagingException {
  // add the addresses
    InternetAddress[] xaddresses;
    String xtext;

    String xcharset = charsetLbl_.getText();

    xtext = fromText_.getText().trim();
    xaddresses = MessageUtilities.encodeAddresses( xtext, xcharset );
    if ( xaddresses.length > 0 )
      message.setFrom( xaddresses[0] );

    xtext = toText_.getText().trim();
    xaddresses = MessageUtilities.encodeAddresses( xtext, xcharset );
    message.addRecipients( Message.RecipientType.TO, xaddresses );

    xtext = ccText_.getText().trim();
    xaddresses = MessageUtilities.encodeAddresses( xtext, xcharset );
    message.addRecipients( Message.RecipientType.CC, xaddresses );

    xtext = bccText_.getText().trim();
    xaddresses = MessageUtilities.encodeAddresses( xtext, xcharset );
    message.addRecipients( Message.RecipientType.BCC, xaddresses );

  // add the subject
    message.setSubject( subjectText_.getText().trim(), xcharset );

  // add the extra headers
    MessageUtilities.addHeaders( message, headers_ );
    MessageUtilities.addHeaders( message, extraHeaders_ );

    return message;
  }

  private MimeMessage
  buildMessage() throws MessagingException {
    MimeMessage xnew = new MimeMessage( ICEMail.getDefaultSession() );
    String xcharset = charsetLbl_.getText();

  // add the content plus any attachments
    String xtext = messageText_.getText();

  // WARNING
  //
  // Remove this translation and SMIME clear signed messages will
  // begin to generate invalid signatures. The problem lies in line
  // termination. The JTextPane likes to give us "\n" terminate lines.
  //
  // However, the mail standards define "\r\n" as the proper line
  // termination sequence. Thus, if we do not convert those "\n"'s
  // to "\r\n", then along the way they will be converted into
  // "\r\n"'s and at that point the signature will become invalid.
  //
  // This is only a problem with "clear signed" messages, since only
  // they can be munged by the intermediate transports.
  //
    try {
      StringBuffer buf = new StringBuffer();
      BufferedReader in = new BufferedReader( new StringReader( xtext ) );

      for ( ; ; ) {
        String inline = in.readLine();
        if ( inline == null )
          break;
        buf.append( inline );
        buf.append( "\r\n" );
      }
      xtext = buf.toString();
    } catch ( IOException ex ) {
      throw new MessagingException( "newline conversion" + ex.getMessage() );
    }

    if ( attachments_.size() > 0 ) {
    // make a multipart message with attachments
      MimeMultipart xmulti = new MimeMultipart();
      if ( xtext != null ) {
        MessageUtilities.attach( xmulti, xtext, xcharset, Part.INLINE );
      }
      MessageUtilities.attach( xmulti, attachments_, xcharset );
      xnew.setContent( xmulti );
    } else {
    // content is simple text
      xnew.setText( xtext, xcharset );
    }

    return xnew;
  }

  private void
  editHeaders() {
    ExtraHeadersDialog dlg = new ExtraHeadersDialog( this, extraHeaders_ );

    dlg.show();

    InternetHeaders xheaders = dlg.getInternetHeaders();
    if ( xheaders.getAllHeaders().hasMoreElements() ) {
      extraHeaders_ = xheaders;
    }
  }

  private void
  pasteAddress( String addrStr ) {
    if ( currentFocus_ != null ) {
      currentFocus_.replaceSelection( addrStr );
    }
  }

// FIXME Move to a utilities class...
  private String
  readFileToString( File f ) {
    String          result = null;

    try {
      StringWriter    out = new StringWriter();
      BufferedReader  in = new BufferedReader( new FileReader( f ) );
      for ( ; ; ) {
        String ln = in.readLine();
        if ( ln == null )
          break;
        out.write( ln );
        out.write( "\n" );
      }
      out.close();
      in.close();

      result = out.toString();
    } catch ( IOException ex ) {
      result = null;
    }

    return result;
  }

  private void
  pasteFile() {
    String title = ICEMail.getBundle().getString( "Compose.title.pastefile" );
    FileDialog dialog = new FileDialog( this, title, FileDialog.LOAD );

    if ( ComposeFrame.LastAttachDir != null )
      dialog.setDirectory( ComposeFrame.LastAttachDir );

    dialog.show();

    String fileName = dialog.getFile();
    String dirName = dialog.getDirectory();
    if ( fileName != null && dirName != null ) {
      ComposeFrame.LastAttachDir = dirName;

      File dirFile = new File( dirName );
      File pasteFile = new File( dirFile, fileName );
      String text = readFileToString( pasteFile );
      if ( text != null ) {
        messageText_.replaceSelection( text );
      }
    }
  }

  private void
  attachItem( Object item, String itemName ) {
    if ( Package.DEBUG && Package.isTraceable( "ComposeFrame" ) ) {
      System.out.println( "ComposeFrame.attachItem(): o-" + item + ", in-[" + itemName + "]" );
    }

    int idx = -1;

    if ( attachments_.size() > 0 )
      idx = attachComboBox_.getSelectedIndex();

    attachments_.insertElementAt( item, idx + 1 );

    attachComboBox_.insertItemAt( itemName, idx + 1 );
    attachComboBox_.setSelectedIndex( idx + 1 );
    attachComboBox_.validate();
    modified_ = true;
  }

  private void
  attachFile() {
    String title = ICEMail.getBundle().getString( "Compose.title.attachfile" );
    FileDialog dialog = new FileDialog( this, title, FileDialog.LOAD );

    if ( ComposeFrame.LastAttachDir != null ) {
      dialog.setDirectory( ComposeFrame.LastAttachDir );
    }

    dialog.show();

    String fileName = dialog.getFile();
    String dirName = dialog.getDirectory();

    if ( fileName != null && dirName != null ) {
      ComposeFrame.LastAttachDir = dirName;

      File dirFile = new File( dirName );
      File attachFile = new File( dirFile, fileName );

      String contentType = FileTypeMap.getDefaultFileTypeMap().getContentType( attachFile );

      String attachName = attachFile.getName().trim();
      //String attachName = attachFile.getName() + " (" + contentType + ")";

      attachItem( attachFile, attachName );
    }
  }

  private void
  detachFile() {
    int idx = attachComboBox_.getSelectedIndex();
    if ( idx >= 0 ) {
      attachments_.removeElementAt( idx );
      attachComboBox_.removeItemAt( idx );
      attachComboBox_.validate();
      modified_ = true;
    }
  }

  private void
  attachSelection( String text ) {
    if ( text != null && text.length() > 0 ) {
      textAttachmentSequence_++;

      String attachName =
        "Text_" + textAttachmentSequence_ + " (text/plain) " + text.length() + " bytes";

      attachItem( text, attachName );
      messageText_.cut();
    }
  }

  private void
  prefixSelection() {
    String selectedText = messageText_.getSelectedText();

    if ( selectedText != null && selectedText.length() > 0 ) {
      StringBuffer buf = new StringBuffer();

      String xprefix = JOptionPane.showInputDialog(
                        ICEMail.getBundle().getString( "Compose.Prefix.dialog.message" ) );

      if ( xprefix != null ) {
        collate( buf, xprefix, selectedText );
        messageText_.replaceSelection( buf.toString() );
        modified_ = true;
      }
    }
  }

  private void
  printMessage() {
    try {
      MimeMessage[] msgs = new MimeMessage[1];

      msgs[0] = buildMessage();
      buildHeader( msgs[0] );
      msgs[0].saveChanges();

      PrintMessageThread pThread = new PrintMessageThread( this, msgs );

      pThread.start();
    } catch ( MessagingException xex ) {
      Object[] xargs = new Object[1];
      xargs[0] = xex.getMessage();
      ComponentFactory.showDialog( ICEMail.getBundle(), "Compose.PrintingException",
                                   0, JOptionPane.ERROR_MESSAGE, xargs );
    }
  }

  private void
  sendMessage() {
    boolean xencrypt = encryptButton_ == null ? false : encryptButton_.current() > 0;
    boolean xsign = signButton_ == null ? false : signButton_.current() > 0;
    boolean xclearsign = signButton_ == null ? false : signButton_.current() == 2;

    try {
      MimeMessage xmessage = buildMessage();
      xmessage.saveChanges();

      if ( xencrypt || xsign ) {
        String charSet = charsetLbl_.getText();
        String fromStr = fromText_.getText().trim();
        InternetAddress[] addrs = MessageUtilities.encodeAddresses( fromStr, charSet );
        xmessage.setFrom( addrs[0] );

        String addrStr;
        Vector addrV = new Vector();
        InternetAddress[] recipAddrs = null;

        if ( xencrypt ) {
          addrStr = toText_.getText().trim();
          recipAddrs = MessageUtilities.encodeAddresses( addrStr, charSet );
          for ( int i = 0 ; i < recipAddrs.length ; ++i ) {
            addrV.addElement( recipAddrs[i] );
          }

          addrStr = ccText_.getText().trim();
          recipAddrs = MessageUtilities.encodeAddresses( addrStr, charSet );
          for ( int i = 0 ; i < recipAddrs.length ; ++i ) {
            addrV.addElement( recipAddrs[i] );
          }

          addrStr = bccText_.getText().trim();
          recipAddrs = MessageUtilities.encodeAddresses( addrStr, charSet );
          for ( int i = 0 ; i < recipAddrs.length ; ++i ) {
            addrV.addElement( recipAddrs[i] );
          }

          recipAddrs = new InternetAddress[ addrV.size() ];
          addrV.copyInto( recipAddrs );
        } else {
          recipAddrs = new InternetAddress[0];
        }

        if ( xencrypt ) {
          xclearsign = false;
        }

        xmessage =
          ICEMail.getSMIMELibrary().buildSMimeMessage( ICEMail.getDefaultSession(),
                                                       xmessage, addrs[0], recipAddrs,
                                                       xsign, xclearsign, xencrypt );
      }

      buildHeader( xmessage );
      xmessage.saveChanges();

      SendMessageThread thread = new SendMessageThread( this, xmessage );
      thread.start();
    } catch ( MissingCertificateException ex ) {
      Object[] xargs = { ex.getMessage() };
      if ( xsign ) {
        ComponentFactory.showDialog( ICEMail.getBundle(), "Compose.MissingCert1",
                                     0, JOptionPane.WARNING_MESSAGE, xargs );
      } else {
        ComponentFactory.showDialog( ICEMail.getBundle(), "Compose.MissingCert2",
                                     0, JOptionPane.WARNING_MESSAGE, xargs );
      }
    } catch ( PasswordCancelException ex ) {
      Object[] xargs = { ex.getMessage() };
      if ( xsign ) {
        ComponentFactory.showDialog( ICEMail.getBundle(), "Compose.MissingPass1",
                                     0, JOptionPane.WARNING_MESSAGE, xargs );
      } else {
        ComponentFactory.showDialog( ICEMail.getBundle(), "Compose.MissingPass2",
                                     0, JOptionPane.WARNING_MESSAGE, xargs );
      }
    } catch ( MessagingException xex ) {
      Object[] xargs = { xex.getMessage() };
      ComponentFactory.showDialog( ICEMail.getBundle(), "Compose.SendingException",
                                   0, JOptionPane.ERROR_MESSAGE, xargs );
    }
  }

  private void
  saveComposition( Message message ) throws MessagingException {
    String xpath = UserProperties.getProperty( "composeFolderPath", null );

    if ( xpath == null ) {
      ComposeConfigDialog xdialog = new ComposeConfigDialog( null );
      xdialog.show();

      xpath = UserProperties.getProperty( "composeFolderPath", null );
    }

    if ( xpath != null && xpath.length() > 0 ) {
      MailUtilities.saveMessage( message, xpath );
    } else {
      String xstring = ICEMail.getBundle().getString( "Exception.noFolderSpecified" );
      throw new MessagingException( xstring );
    }

  }

  private void
  collate( StringBuffer buf, String prefix, String text ) {
    boolean wasNewline = false;
    StringTokenizer toker = new StringTokenizer( text, "\n", true );
    int count = toker.countTokens();

    for ( int i = 0 ; i < count ; ++i ) {
      String textLine;
      try {
        textLine = toker.nextToken();
      } catch ( NoSuchElementException ex ) {
        break;
      }

      if ( ! textLine.equals( "\n" ) ) {
        buf.append( prefix );
        buf.append( textLine );
        wasNewline = false;
      } else {
        if ( wasNewline ) {
          buf.append( prefix );
        }
        buf.append( "\n" );
        wasNewline = true;
      }
    }
  }

  private void
  setSignature( String sig ) {
    int selEnd = messageText_.getSelectionEnd();
    int selStart = messageText_.getSelectionStart();

    if ( signature_ != null ) {
      String msgText = messageText_.getText();

      int index = -1;
      if ( signature_.length() > 0 ) {
        index = msgText.indexOf( signature_ );
      }

      if ( index != -1 ) {
        selStart = index;
        messageText_.setCaretPosition( index );
        messageText_.moveCaretPosition( index + signature_.length() );
      }
    }

    signature_ = sig;
    messageText_.replaceSelection( sig );

    selEnd = selStart + sig.length();

    messageText_.setCaretPosition( selEnd );
    messageText_.moveCaretPosition( selEnd );

    modified_ = true;
  }

  private void
  establishContents() {
    Container content = getContentPane();
    ((JPanel) content).setDoubleBuffered( true );

    content.setLayout( new GridBagLayout() );

    JToolBar xtoolBar = new JToolBar();
    xtoolBar.setFloatable( false );
    AWTUtilities.constrain( content, xtoolBar, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.NORTHWEST, 0, 0, 1, 1, 0.5, 0.0 );

    sendButton_ = ComponentFactory.getButton( ICEMail.getBundle(), "Compose.Send",
                                              actionListener_, focusListener_ );
    xtoolBar.add( sendButton_ );

    xtoolBar.addSeparator();

    printButton_ = ComponentFactory.getButton( ICEMail.getBundle(), "Compose.Print",
                                               actionListener_, focusListener_ );
    xtoolBar.add( printButton_ );

    xtoolBar.addSeparator();

    priorityButton_ = ComponentFactory.getRotateButton( ICEMail.getBundle(), "Compose.Priority",
                                                        actionListener_, focusListener_ );
    xtoolBar.add( priorityButton_ );

    xtoolBar.addSeparator();

    if ( ICEMail.getSMIMELibrary() != null ) {
      encryptButton_ = ComponentFactory.getRotateButton( ICEMail.getBundle(), "Compose.Encrypt",
                                                         actionListener_, focusListener_ );
      xtoolBar.add( encryptButton_ );

      signButton_ = ComponentFactory.getRotateButton( ICEMail.getBundle(), "Compose.Sign",
                                                      actionListener_, focusListener_ );
      xtoolBar.add( signButton_ );
    }

    JPanel comboPan = new JPanel();
    comboPan.setLayout( new GridBagLayout() );
    AWTUtilities.constrain( content, comboPan, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.NORTHEAST, 1, 0, 1, 1, 0.5, 0.0 );

    JToolBar attachTools = new JToolBar();
    attachTools.setFloatable( false );
    AWTUtilities.constrain( comboPan, attachTools, GridBagConstraints.NONE,
    GridBagConstraints.EAST, 0, 0, 1, 1, 0.0, 0.0 );

    attachButton_ = ComponentFactory.getButton( ICEMail.getBundle(), "Compose.Attach",
                                                actionListener_, focusListener_ );
    attachTools.add( attachButton_ );

    detachButton_ = ComponentFactory.getButton( ICEMail.getBundle(), "Compose.Detach",
                                                actionListener_, focusListener_ );
    attachTools.add( detachButton_ );

    attachComboBox_ = new JComboBox();
    attachComboBox_.addFocusListener( focusListener_ );
    AWTUtilities.constrain( comboPan, attachComboBox_, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.WEST, 1, 0, 1, 1, 1.0, 0.0 );

    JPanel xpanel = new JPanel();

    xpanel.addFocusListener( focusListener_ );
    xpanel.setLayout( new GridBagLayout() );
    xpanel.setBorder( JFCUtilities.StandardBorder );

    AWTUtilities.constrain( content, xpanel, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.NORTHWEST, 0, 1, 2, 1, 1.0, 0.0 );

    JLabel  lbl;
    int     row = 0;

    lbl = ComponentFactory.getLabel( ICEMail.getBundle(), "Compose.From" );
    AWTUtilities.constrain( xpanel, lbl, GridBagConstraints.NONE,
                            GridBagConstraints.NORTHWEST, 0, row, 1, 1, 0.0, 0.0 );
    fromText_ = ComponentFactory.getTextField( ICEMail.getBundle(), "Compose.From", focusListener_ );
    fromText_.addActionListener( new AddressFieldActionListener( ICEMail.getAddressBook() ) );
    AWTUtilities.constrain( xpanel, fromText_, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.NORTHWEST, 1, row++, 1, 1, 1.0, 0.0 );

    lbl = ComponentFactory.getLabel( ICEMail.getBundle(), "Compose.To" );
    AWTUtilities.constrain( xpanel, lbl, GridBagConstraints.NONE,
                            GridBagConstraints.NORTHWEST, 0, row, 1, 1, 0.0, 0.0 );
    toText_ = ComponentFactory.getTextField( ICEMail.getBundle(), "Compose.To", focusListener_ );
    toText_.addActionListener( new AddressFieldActionListener( ICEMail.getAddressBook() ) );
    AWTUtilities.constrain( xpanel, toText_, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.NORTHWEST, 1, row++, 1, 1, 1.0, 0.0 );

    lbl = ComponentFactory.getLabel( ICEMail.getBundle(), "Compose.Cc" );
    AWTUtilities.constrain( xpanel, lbl, GridBagConstraints.NONE,
                            GridBagConstraints.NORTHWEST, 0, row, 1, 1, 0.0, 0.0 );
    ccText_ = ComponentFactory.getTextField( ICEMail.getBundle(), "Compose.Cc", focusListener_ );
    ccText_.addActionListener( new AddressFieldActionListener( ICEMail.getAddressBook() ) );
    AWTUtilities.constrain( xpanel, ccText_, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.NORTHWEST, 1, row++, 1, 1, 1.0, 0.0 );

    lbl = ComponentFactory.getLabel( ICEMail.getBundle(), "Compose.Bcc" );
    AWTUtilities.constrain( xpanel, lbl, GridBagConstraints.NONE,
                            GridBagConstraints.NORTHWEST, 0, row, 1, 1, 0.0, 0.0 );
    bccText_ = ComponentFactory.getTextField( ICEMail.getBundle(), "Compose.Bcc", focusListener_ );
    bccText_.addActionListener( new AddressFieldActionListener( ICEMail.getAddressBook() ) );
    AWTUtilities.constrain( xpanel, bccText_, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.NORTHWEST, 1, row++, 1, 1, 1.0, 0.0 );

    lbl = ComponentFactory.getLabel( ICEMail.getBundle(), "Compose.Subject" );
    AWTUtilities.constrain( xpanel, lbl, GridBagConstraints.NONE,
                            GridBagConstraints.NORTHWEST, 0, row, 1, 1, 0.0, 0.0 );
    subjectText_ = ComponentFactory.getTextField( ICEMail.getBundle(), "Compose.Subject",
                                                  focusListener_ );
    AWTUtilities.constrain( xpanel, subjectText_, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.NORTHWEST, 1, row++, 1, 1, 1.0, 0.0 );

    DefaultStyledDocument doc = new DefaultStyledDocument();

    messageUndo_ = new UndoManager();
    doc.addUndoableEditListener( messageUndo_ );
    messageText_ = new JTextPane( doc );

    JScrollPane xscrollPane = new JScrollPane();
    xscrollPane.getViewport().add( messageText_ );

    messageText_.getCaret().setBlinkRate( 500 );
    messageText_.addFocusListener( focusListener_ );

    AWTUtilities.constrain( content, xscrollPane, GridBagConstraints.BOTH,
                            GridBagConstraints.CENTER, 0, 2, 2, 1, 2.0, 1.0 );

    JPanel statusPan = new JPanel();
    statusPan.setLayout( new GridBagLayout() );

    AWTUtilities.constrain( content, statusPan, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.SOUTHWEST, 0, 3, 1, 1, 1.0, 0.0 );

    charsetLbl_ = new JLabel( (String) charsets_.firstElement() );

    charsetMenu_ = new JPopupMenu();
    for ( int i = 0, sz = charsets_.size() ; i < sz ; ++i ) {
      String cs = (String) charsets_.elementAt(i);
      Action act = new IAbstractAction( cs );
      charsetMenu_.add( act );
    }

    charsetLbl_.addMouseListener( new IMouseAdapter( charsetMenu_ ) );

    AWTUtilities.constrain( statusPan, charsetLbl_, GridBagConstraints.HORIZONTAL,
                            GridBagConstraints.SOUTHWEST, 0, 0, 1, 1, 1.0, 0.0 );

    doc.addUndoableEditListener( new IUndoableEditListener() );
  }

  private void
  establishMenuBar() {
    menuBar_ = new JMenuBar();

    menuBar_.add( ComponentFactory.getMenuAndItems( ICEMail.getBundle(),
                                                    "Compose.Message",
                                                    actionListener_ ) );
    menuBar_.add( ComponentFactory.getMenuAndItems( ICEMail.getBundle(),
                                                    "Compose.Edit",
                                                    actionListener_ ) );

    establishSignatures();
    establishAddresses();

    setJMenuBar( menuBar_ );
  }

  private void
  establishSignatures() {
    JMenuItem item;

    String sigListStr = UserProperties.getProperty( "signatureList", null );

    JMenu menu = ComponentFactory.getMenuAndItems( ICEMail.getBundle(), "Compose.Sign",
                                                   actionListener_ );

    menuBar_.add( menu );

    menu.addSeparator();

    if ( sigListStr == null )
      return;

    String[] sigList = StringUtilities.splitString( sigListStr, ":" );

    for ( int i = 0 ; i < sigList.length ; ++i ) {
      if ( sigList[i].equals( "default" ) )
        continue;

      String sigStr = UserProperties.getProperty( ("signature." + sigList[i]), null );

      if ( sigStr == null )
        continue;

      item = new JMenuItem( sigList[i] );
      item.addActionListener( actionListener_ );
      item.setActionCommand( "SIGN:" + sigList[i] );
      menu.add( item );
    }
  }

  private void
  establishAddresses() {
    JMenuItem item;
    String addrListStr = UserProperties.getProperty( "addressList", null );

// begin additions for LDAP lookup (mcs@netscape.com)
	/*
	 * If an LDAP directory is configured, include a
	 * "Directory Lookup" item in the Addresses menu.
	 */
	String ldapURLStr = UserProperties.getProperty( "ldapDirectory", null );
	if ( ldapURLStr != null ) {
	  try {
		ldapDirectoryURL = new LDAPUrl( ldapURLStr );
	  } catch ( MalformedURLException e ) {
		JOptionPane.showMessageDialog( null,
				"Malformed LDAP URL: " + ldapURLStr );
	  }
	}

    String[] addrList = {};
	if ( addrListStr != null )
	  addrList = StringUtilities.splitString( addrListStr, ":" );
    if ( addrList.length < 1 && ldapDirectoryURL == null )
      return;

    JMenu menu = ComponentFactory.getMenu( ICEMail.getBundle(),
			"Compose.Address" );

    menuBar_.add( menu );

	if ( ldapDirectoryURL != null ) {
	  item = new JMenuItem( "Directory Lookup" );
	  item.addActionListener( actionListener_ );
	  item.setActionCommand( "LDAP:LOOKUP" );
	  item.setAccelerator( KeyStroke.getKeyStroke( 'L',
			Event.CTRL_MASK ) );
	  menu.add( item );
	  if ( addrList.length > 0 ) menu.addSeparator();
	}


// end of additions by mcs@netscape.com

    for ( int i = 0 ; i < addrList.length ; ++i ) {
      String addrStr = UserProperties.getProperty( ("address." + addrList[i]), null );

      if ( addrStr == null )
        continue;

      item = new JMenuItem( addrList[i] );
      item.addActionListener( actionListener_ );
      item.setActionCommand( "ADDR:" + addrList[i] );
      if ( i < 10 )
        item.setAccelerator( KeyStroke.getKeyStroke( ComposeFrame.AccKeys[i], Event.CTRL_MASK ) );
      menu.add( item );
    }
  }

// ............................................................
// these routines setup the contents of the composition

  private void
  setMessage( String body ) {
    try {
      Message xnew = null;

      if ( originalMessage_ != null ) {
        String xprefix = null;
        String xenclosure = null;
        boolean xuseattachments = false;

        Vector descV = new Vector();
        Vector attachV = new Vector();
        StringBuffer bodyBuf = new StringBuffer( 2046 );

        MessageUtilities.decodeContent( originalMessage_, bodyBuf, attachV, descV );

      // create or use the appropriate message
        switch ( state_ ) {
        case FORWARDING:
          xprefix = getForwardPrefix( originalMessage_ );
          xenclosure = getForwardEnclosure( originalMessage_ );
          xuseattachments =
            UserProperties.getProperty( Configuration.P_COMPOSE_FORWARD_ATTACHMENTS, true );
          xnew = MessageUtilities.createForward( originalMessage_ );
          break;

        case REPLYING:
          xprefix = getReplyPrefix( originalMessage_ );
          xenclosure = getReplyEnclosure( originalMessage_ );
          xnew = MessageUtilities.createReply( originalMessage_, "", false );
          break;

        case REPLYINGTOALL:
          xprefix = getReplyPrefix( originalMessage_ );
          xenclosure = getReplyEnclosure( originalMessage_ );
          xnew = MessageUtilities.createReply( originalMessage_, "", true );
          break;

        default:
          xuseattachments = true;
          xnew = originalMessage_;
          break;
        }

      // now convert the resulting message to the display
        fromText_.setText( MessageUtilities.decodeFrom( xnew ) );

        toText_.setText( MessageUtilities.decodeAddresses( xnew, Message.RecipientType.TO ) );

        ccText_.setText( MessageUtilities.decodeAddresses( xnew, Message.RecipientType.CC ) );

        bccText_.setText( MessageUtilities.decodeAddresses( xnew, Message.RecipientType.BCC ) );

        subjectText_.setText( xnew.getSubject() );

        if ( body != null ) {
          messageText_.setText( body );
        } else {
          messageText_.setText( bodyBuf.toString() );
        }

        if ( xuseattachments ) {
          for ( int ai = 0, sz = attachV.size() ; ai < sz ; ai++ ) {
          // attachment
            Object ap = attachV.elementAt(ai);
            attachments_.addElement( ap );
          // name
            String ad = (String) descV.elementAt(ai);
            attachComboBox_.addItem( ad );
          }
        }

        headers_ = MessageUtilities.getHeaders( xnew );
        headers_.removeHeader( "from" );
        headers_.removeHeader( "to" );
        headers_.removeHeader( "cc" );
        headers_.removeHeader( "bcc" );
        headers_.removeHeader( "subject" );
        headers_.removeHeader( "content-type" );

      // apply any adornments
        if ( xprefix != null ) {
          prefixMessageText( xprefix );
        }

        if ( xenclosure != null ) {
          encloseMessageText( xenclosure );
        }

        if ( MessageUtilities.isHighPriority( originalMessage_ ) ) {
          priorityButton_.rotate( 1 );
        } else if ( MessageUtilities.isLowPriority( originalMessage_ ) ) {
          priorityButton_.rotate( 3 );
        }
      } else {
      // no supplied message, just provide a
      // from address to start with
        fromText_.setText( MessageUtilities.getDefaultFromString() );
      }

    // add the custom headers ( X-Mailer, Organization... )
      String orgStr = UserProperties.getProperty( Configuration.P_ORGANIZATION, null );

      String mailerStr =
        "ICEMail (rel " + ICEMail.getVersionString() + ") <http://www.icemail.org/>";

      headers_.setHeader( "X-Mailer", mailerStr );

      if ( orgStr != null && orgStr.length() > 0 ) {
        headers_.setHeader( "Organization", orgStr );
      }
    } catch ( MessagingException xex ) {
      System.out.println( "ComposeFrame.setMessage() failed: " + xex.toString() );
    }
  }

  private String
  formatString( String format, Message msg ) {
  // create the arguments from the message contents
    Object[] xargs = new Object[4];
    xargs[0] = "\n";
    try {
      xargs[1] = MessageUtilities.decodeFrom( msg );
    } catch (Exception xex) {
      xargs[1] = "unknown";
    }
    try {
      xargs[2] = msg.getSentDate();
    } catch (Exception xex) {
      xargs[2] = new Date();
    }
    try {
      xargs[3] = msg.getSubject();
    } catch (Exception xex) {
      xargs[3] = "";
    }
  // replace the arguments in the given format
    return MessageFormat.format( format, xargs );
  }

  private String
  getReplyPrefix( Message msg ) {
    String xresult =
      UserProperties.getProperty( Configuration.P_COMPOSE_REPLY_PREFIX, null );

    if ( xresult != null && xresult.length() > 0 ) {
      xresult = formatString( xresult, msg );
    } else {
      xresult = null;
    }
    return xresult;
  }

  private String
  getReplyEnclosure( Message msg ) {
    String xresult =
      UserProperties.getProperty( Configuration.P_COMPOSE_REPLY_ENCLOSURE, null );

    if ( xresult != null && xresult.length() > 0 ) {
      xresult = formatString( xresult, msg );
    } else {
      xresult = null;
    }
    return xresult;
  }

  private String
  getForwardPrefix( Message msg ) {
    String xresult =
      UserProperties.getProperty( Configuration.P_COMPOSE_FORWARD_PREFIX, null );

    if ( xresult != null && xresult.length() > 0 ) {
      xresult = formatString( xresult, msg );
    } else {
      xresult = null;
    }
    return xresult;
  }

  private String
  getForwardEnclosure( Message msg ) {
    String xresult =
      UserProperties.getProperty( Configuration.P_COMPOSE_FORWARD_ENCLOSURE, null );

    if ( xresult != null && xresult.length() > 0 ) {
      xresult = formatString( xresult, msg );
    } else {
      xresult = null;
    }
    return xresult;
  }

  private void
  prefixMessageText( String prefix ) {
    String xcontents = messageText_.getText();

    if ( xcontents == null || xcontents.length() == 0 )
      return;

    try {
      BufferedReader xreader =
        new BufferedReader( new StringReader( xcontents ) );
      StringBuffer xnew = new StringBuffer( xcontents.length() );

      for ( String xline = null; ( xline = xreader.readLine() ) != null ; ) {
        xnew.append( prefix + xline + '\n' );
      }
      messageText_.setText( xnew.toString() );
    } catch( IOException xex ) { }
  }

  private void
  encloseMessageText( String enclosure ) {
    String xreplacement = '\n' + enclosure + "\n\n";

  // prefix the message with the enclosure
    messageText_.setCaretPosition( 0 );
    messageText_.moveCaretPosition( 0 );
    messageText_.replaceSelection( xreplacement );

  // postfix the message with the enclosure
    int end = messageText_.getDocument().getEndPosition().getOffset();
    messageText_.setCaretPosition( end - 1 );
    messageText_.moveCaretPosition( end - 1 );
    messageText_.replaceSelection( xreplacement );
  }

//............................................................

  private class IActionListener
    implements ActionListener
  {
    /**
     * Invoked when an action occurs.
     * <p>
     * Implementation of java.awt.event.ActionListener.actionPerformed()
     *
     * @param event the action event that occured
     */
    public void
    actionPerformed( ActionEvent event ) {
    // FIXME determine which commands need to call the following:
    // getInputContext().endComposition();

      String command = event.getActionCommand();

      if ( command.equals( "CLOSE" ) ) {
        closeWindow();
      } else if ( command.equals( sendButton_.getActionCommand() ) ) {
        sendMessage();
      } else if ( command.equals( printButton_.getActionCommand() ) ) {
        printMessage();
      } else if ( command.startsWith( "ADDR:" ) ) {
        String addrStr = null;
        String addrName = command.substring( 5 );

        addrStr = UserProperties.getProperty( ("address." + addrName), null );

        if ( addrStr != null ) {
          pasteAddress( addrStr );
        }
      } else if ( command.startsWith( "SIGN:" ) ) {
        String sigStr = null;
        String sigName = command.substring( 5 );

        if ( sigName.equals( "DEFAULT" ) ) {
          sigStr = UserProperties.getProperty( "signature.default", null );
        } else if ( sigName.equals( "NONE" ) ) {
          sigStr = "";
        } else {
          sigStr = UserProperties.getProperty( ("signature." + sigName), null );
        }

        if ( sigStr != null ) {
          setSignature( sigStr );
        }
      } else if ( command.equals( "HEADERS" ) ) {
        editHeaders();
      } else if ( command.equals( attachButton_.getActionCommand() ) ) {
        String text = messageText_.getSelectedText();
        if ( text != null && text.length() > 0 ) {
          attachSelection( text );
        } else {
          attachFile();
        }
      } else if ( command.equals( "PREFIX" ) ) {
        prefixSelection();
      } else if ( command.equals( detachButton_.getActionCommand() ) ) {
        detachFile();
      } else if ( command.equals( "CUT" ) ) {
        if ( currentFocus_ != null ) {
          currentFocus_.cut();
        }
      } else if ( command.equals( "COPY" ) ) {
        if ( currentFocus_ != null ) {
          currentFocus_.copy();
        }
      } else if ( command.equals( "PASTE" ) ) {
        if ( currentFocus_ != null ) {
          currentFocus_.paste();
        }
      } else if ( command.equals( "PASTEFILE" ) ) {
        if ( messageText_ != null ) {
          pasteFile();
        }
      } else if ( command.equals( "UNDO" ) ) {
        if ( messageUndo_ != null && messageUndo_.canUndoOrRedo() ) {
          messageUndo_.undoOrRedo();
        }
      } else if ( command.equals( "CHARSET" ) ) {
        int xitem = charsets_.indexOf( charsetLbl_.getText() ) + 1;

        if ( xitem >= charsets_.size() ) {
          xitem = 0;
        }
        charsetLbl_.setText( (String)charsets_.elementAt( xitem ) );
      } else if ( command.equals( "SIGNON" ) ) {
      } else if ( command.equals( "SIGNCLEAR" ) ) {
      } else if ( command.equals( "SIGNOFF" ) ) {
      } else if ( command.equals( "ENCRYPTON" ) ) {
      } else if ( command.equals( "ENCRYPTOFF" ) ) {
      } else if ( command.equals( "HIGHPRIORITY" ) ) {
        headers_.setHeader( "X-Priority", "1" );
        headers_.setHeader( "Importance", "high" );
      } else if ( command.equals( "NORMALPRIORITY" ) ) {
        headers_.removeHeader( "X-Priority" );
        headers_.removeHeader( "Importance" );
      } else if ( command.equals( "LOWPRIORITY" ) ) {
        headers_.setHeader( "Importance", "low" );
        headers_.setHeader( "X-Priority", "5" );
// begin additions for LDAP lookup (mcs@netscape.com)
      } else if ( command.equals( "LDAP:LOOKUP" ) ) {
		ldapLookup();
// end of additions by mcs@netscape.com
      } else {
        System.err.println( "ComposeFrame.actionPerformed() - unknown command '" +
                            command + "'" );
      }
    }
  }

//............................................................

    private class IWindowDisposer
      implements Runnable
    {
      public void run() {
        setVisible( false );
        setEnabled( false );
      }
    }

//............................................................

    private class IWindowAdapter
      extends WindowAdapter
    {
      public void
      windowClosing( WindowEvent e )
      {
        //System.out.println( "windowClosing(e)" );
        closeWindow();
      }
    }

//............................................................

    private class IFocusListener
      implements FocusListener
    {
      /**
       * <p>
       * Implementation of java.awt.event.FocusListener.focusLost()
       */
      public void
      focusLost( FocusEvent event ) { }

      /**
       * <p>
       * Implementation of java.awt.event.FocusListener.focusGained()
       */
      public void
      focusGained( FocusEvent event ) {
        Object comp = event.getSource();

        if ( comp instanceof JTextField || comp instanceof JTextPane ) {
          currentFocus_ = (JTextComponent)event.getSource();
        } else {
          currentFocus_ = null;
        }
      }
    }

//............................................................

    private class IUndoableEditListener
      implements UndoableEditListener
    {
      public void
      undoableEditHappened( UndoableEditEvent event ) {
        modified_ = true;
      }
    }

//............................................................

    private class IMouseAdapter
      extends MouseAdapter
    {
      private boolean       isPopupClick = false;
      private JPopupMenu    popup = null;

      public
      IMouseAdapter( JPopupMenu popup ) {
        super();
        this.popup = popup;
      }

      public void
      mousePressed( MouseEvent event ) {
        this.isPopupClick = false;

        if ( event.isPopupTrigger() ) {
          this.isPopupClick = true;
          Point pt =
            JFCUtilities.computePopupLocation( event, charsetLbl_, popup );
          this.popup.show( charsetLbl_, pt.x, pt.y );
        }
      }

      public void
      mouseReleased( MouseEvent event ) {
        if ( this.isPopupClick ) return;

        if ( event.isPopupTrigger() ) {
          this.isPopupClick = true;
          Point pt =
            JFCUtilities.computePopupLocation( event, charsetLbl_, popup );
          this.popup.show( charsetLbl_, pt.x, pt.y );
        }
      }

      public void
      mouseClicked( MouseEvent event ) {
        if ( this.isPopupClick ) {
          this.isPopupClick = false;
          return;
        }
        if ( event.getClickCount() == 2 ) {
          processDoubleClick();
        }
      }

      private void
      processDoubleClick() { }
    }

//............................................................

    private class IAbstractAction
    extends AbstractAction
    {
      private String charset_;

      public
      IAbstractAction( String charset ) {
        super( charset );
        charset_ = charset;
      }

      public void
      actionPerformed( ActionEvent  evt ) {
        int xitem = charsets_.indexOf( charset_ ) + 1;
        if ( xitem >= charsets_.size() )
          xitem = 0;
        charsetLbl_.setText( charset_ );
      }
    }

//............................................................
	/*
	 * ldapLookup: grab the contents of the current text entry field,
	 * look it up in the configured LDAP directory, and replace the
	 * original text with DISPLAYNAME <MAIL> or CN <MAIL>.
	 */
	private void
	ldapLookup() {
	  String			ldAttrs[] = { "cn", "displayName", "mail" };
	  LDAPConnection	ld = null;

	  // Get text from the active text entry field
	  if ( currentFocus_ == null ) {
		return;
	  }
	  String textToLookup = currentFocus_.getSelectedText();
	  boolean useSelection = ( textToLookup != null );
	  if ( !useSelection ) {
		textToLookup = currentFocus_.getText();
	  }
	  if ( textToLookup.length() <= 0 ) {
		return;
	  }

	  // Create filter and execute an LDAP search
	  String ldFilter = "(&(objectClass=person)";
	  if ( textToLookup.indexOf( "=" ) != -1 ) {
		ldFilter += "(" + textToLookup + "))";
	  } else {
		ldFilter += "(|(cn=" + textToLookup + ")"
				+ "(uid=" + textToLookup + ")"
				+ "(sn=" + textToLookup + ")))";
	  }

	  try {
	    ld = new LDAPConnection();
		ld.connect( ldapDirectoryURL.getHost(),
				ldapDirectoryURL.getPort() );

		LDAPSearchConstraints cons = new LDAPSearchConstraints();
		int sizelimit = UserProperties.getProperty( "ldapSizeLimit", 100 );
		int timelimit = UserProperties.getProperty( "ldapTimeLimit", 60 );
		cons.setMaxResults( sizelimit );		// entries
		cons.setServerTimeLimit( timelimit );	// seconds

		LDAPSearchResults res = ld.search(
				ldapDirectoryURL.getDN(),
				LDAPConnection.SCOPE_SUB, ldFilter, ldAttrs, false,
				cons );

		// Process the search results
		int			count = 0;
		LDAPEntry	entry = null;

		while ( res.hasMoreElements()) {
		  try {
			entry = res.next();
			++count;
		  } catch ( LDAPException e ) {
			JOptionPane.showMessageDialog( null,
					"LDAP res.next error: " + e.errorCodeToString());
		  }
		}

		if ( count <= 0 ) {
		  JOptionPane.showMessageDialog( null,
				  "No entries matched " + textToLookup );
		} else if ( count > 1 ) {
		  JOptionPane.showMessageDialog( null,
				  count + " entries matched " + textToLookup );
		} else {
		  /*
		   * Exactly one match... replace contents of current field
		   * (or current selection) with:
		   *	displayName <mail>
		   * or:	
		   *	cn <mail>
		   */
		  String		newAddress;
		  Enumeration	enum;
		  LDAPAttribute	attr = entry.getAttribute( "displayName" );
		  if ( attr == null ) {
			  attr = entry.getAttribute( "cn" );
		  }
		  if ( attr != null ) {
			  enum = attr.getStringValues();
			  newAddress = (String)enum.nextElement() + " ";
		  } else {
			  newAddress = "";
		  }

		  attr = entry.getAttribute( "mail" );
		  enum = attr.getStringValues();
		  newAddress += "<" + (String)enum.nextElement() + ">";
		  if ( useSelection ) {
			currentFocus_.replaceSelection( newAddress );
		  } else {
			currentFocus_.setText( newAddress );
		  }
		}

	  } catch ( LDAPException e ) {
		JOptionPane.showMessageDialog( null,
				"LDAP connect or search error: "
			    + e.errorCodeToString());
	  }

	  // Clean up -- close the LDAP connection
	  if ( ld != null && ld.isConnected()) {
		try {
		  ld.disconnect();
		} catch ( LDAPException e ) {
		  /* ignore */
		}
	  }
	}
// end of additions by mcs@netscape.com

}
