#!/usr/local/bin/perl
#
# lfingerd -- a Perl 5 script that implements the server side
#   of the BSD finger protocol using an LDAP server as its
#   database of user information.  Also returns AIM
#	online/offline status if Netscape Directory Server 6 or
#	later is used.
#
# From the 2nd Edition of the book:
#	"Understanding and Deploying LDAP Directory Services"
#	by Timothy A. Howes, Mark C. Smith, and Gordon S. Good.
#
# Requires: PerLDAP
#

use Mozilla::LDAP::Conn;

# Banner:
$banner = "Example.Com LDAP Finger Service";

# LDAP server information:
$ldapbase = "dc=example,dc=com";
#$ldaphost = "ldap.example.com";
#$ldapport = "389";

# Constants:
$unknown = "???";
$separator = "\n";
@attrlist = ( "cn", "uid", "mail", "departmentNumber",
    "telephoneNumber", "roomNumber", "postalAddress",
    "ou", "title", "displayName", "description",
    "labeledURI", "nsAIMID", "nsAIMStatusText" );

# Start of main:
$| = 1;		# enable autoflush of output upon print
print "$banner\n\n";
$| = 0;		# disable autoflush

# grab query string and chop off newline and return characters
$query = <STDIN>;
chop $query;
if ( $query =~ /\r$/ ) {
	chop $query;
}

# form a filter using the query string
if ( $query =~ /.*=.*/ ) {
	$filter = "(&(objectClass=person)($query))";
} else {
	$filter = "(&(objectClass=person)"
			. "(|(sn=$query)(cn=$query)(uid=$query)))";
}


# open an anonymous connection to the LDAP server
$ldap = new Mozilla::LDAP::Conn( $ldaphost, $ldapport );
die "Unable to connect to server at "
		. "ldap://$ldaphost:$ldapport\n" unless $ldap;

# search the directory
$entry = $ldap->search( $ldapbase, "subtree", $filter,
		0, @attrlist );

# display all the results
if ( $entry ) {
	$count = 0;
	do {
		displayEntry( $entry );
		$entry = $ldap->nextEntry;
		++$count;
	} while ( $entry );

	if ( $count > 1 ) {
		print $count, " entries matched '$query'\n";
	} else {
		print "One entry matched '$query'\n";
	}

} else {
	print "No entries matched '$query'\n";
}

# Close LDAP connection and clean up
$ldap->close;

# End of main.


# Start of displayEntry:
sub
displayEntry {
	local( $entry ) = @_;
	local( $value, @attrs );

	@attrs = ( "displayName", "cn" );
	printf( "Login name:  %-8s             "
			. "            In real life: %s\n",
			getSimpleValue( $entry, "uid" ),
			getFirstValue( $entry, *attrs ));

	printf( "E-Mail:      %-32s Phone: %s\n",
			getSimpleValue( $entry, "mail" ),
			getSimpleValue( $entry, "telephoneNumber" ));

	@attrs = ( "ou", "departmentNumber" );
	printf( "Department:  %-32s Room:  %s\n",
			getFirstValue( $entry, *attrs ),
			getSimpleValue( $entry, "roomNumber" ));

	$value = getSimpleValue( $entry, "nsAIMID" );
	if ( $value ne $unknown ) {
		printf( "AIM ID:      %s (status: %s)\n", $value,
				getSimpleValue( $entry, "nsAIMStatusText" ));
	}

	$value = getSimpleValue( $entry, "postalAddress" );
	if ( $value ne $unknown ) {
		print "Address:     $value\n";
	}

	$value = getSimpleValue( $entry, "title" );
	if ( $value ne $unknown ) {
		print "Title:       $value\n";
	}

	$value = getSimpleValue( $entry, "description" );
	if ( $value ne $unknown ) {
		print "Description: $value\n";
	}

	displayAllValues( $entry, "labeledURI", "URL:         " );

	print $separator;
}
# End of displayEntry.


# Start of displayAllValues:
sub
displayAllValues {
	local( $entry, $attr, $prefix ) = @_;
	local( $value );

	if ( $entry->{$attr} ) {
		foreach $value (@{$entry->{$attr}}) {
			print $prefix, $value, "\n";
		}
	}
}
# End of displayAllValues.

# Start of getSimpleValue:
sub
getSimpleValue {
	local( $entry, $attr ) = @_;
	local( $value );

	if ( $entry->{$attr} ) {
		$value = $entry->{$attr}[0];
	} else {
		$value = $unknown;
	}

	$value;
}
# End of getSimpleValue.

# Start of getFirstValue:
sub
getFirstValue {
	local( $entry, *attrs ) = @_;
	local( $a );
	local( $value );

	$value = $unknown;

	foreach $a (@attrs) {
		$value = getSimpleValue( $entry, $a );
		last if ( $value ne $unknown );
	}

	$value;
}
# End of getFirstValue.
